/*
** Module   : ScrollBar.java
** Abstract : generic scroll bar. 
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141015 Notify scroll listeners only if position is in range.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150509 Allow the scrollbar to use double values for step/scrolling.
** 005 EVL 20150512 Need to introduce one more parameter to correctly calculate thumb size in
**                  GUI mode.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A generic scroll bar. 
 */
public abstract class ScrollBar<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<BaseConfig>,
           ScrollListener
           
{   
   /**
    * Container bars. Containers having scroll bars could have:
    * 
    * NONE       - No scroll bars.
    * HORIZONTAL - Only horizontal scroll bar.
    * VERTICAL   - Only vertical scroll bar.
    * BOTH       - Both horizontal and vertical scroll bars.
    */
   public enum Bars
   {
      NONE,
      HORIZONTAL,
      VERTICAL,
      BOTH
   }
   
   /**
    * Scroll bar orientation enumeration. Scroll bars are of two kinds horizontal or vertical. 
    */
   public enum Orientation
   {
      HORIZONTAL,
      VERTICAL
   }

   /**
    * Scroll bar buttons position. For a vertical scroll bar buttons positions are as follow:
    * LEFT  button is at the TOP. 
    * RIGHT button is at the BOTTOM.
    */
   public enum Position
   {
      LEFT,
      RIGHT
   }
   
   /** Scroll bar orientation */
   protected Orientation orientation;

   /** Scroll minimum value */
   protected double min = 0;
   
   /** Scroll maximum value */
   protected double max = 0;
            
   /** Current position */
   protected double position = 0;
   
   /** Scroll step */
   protected double step = 1;

   /** Scroll visible area range, default is 5. */
   protected double visibleRange = 5.0;
            
   /** Storage for listeners. */
   protected ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /** Widget configuration */
   protected BaseConfig config;
   
   /**
    * Default constructor. Build a horizontal scroll bar.
    */
   public ScrollBar(WidgetId id)      
   {
      this(id, Orientation.HORIZONTAL);
   }

   /**
    * Constructor.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    */
   public ScrollBar(WidgetId id, Orientation orientation)
   {
      super();
      this.orientation = orientation;
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Initialize scroll bar parameters.
    */
   public void clear()
   {
      min = 0;
      max = 0;
      position = 0;
   }
   
   /**
    * Return scroll bar range.
    */
   public double range()
   {
      return (max - min);
   }

   /**
    * Add scroll listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   public void addScrollListener(ScrollListener listener)
   {
      scrollListeners.add(listener);
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Get minimum value.
    * 
    * @return  see above.
    */
   public double getMin()
   {
      return min;
   }

   /**
    * Set minimum value.
    * 
    * @param   min
    *          New value.
    */
   public void setMin(double min)
   {
      this.min = min;
   }

   /**
    * Return maximum value.
    * 
    * @return  see above.
    */
   public double getMax()
   {
      return max;
   }

   /**
    * Set maximum value.
    * 
    * @param   max
    *          New maximum value.
    */
   public void setMax(double max)
   {
      this.max = max;
   }

   /**
    * Get current position.
    * 
    * @return  see above.
    */
   public double getPosition()
   {
      return position;
   }

   /**
    * Set current position. Position MUST be in range.
    * 
    * @param   position
    *          New position.
    */
   public void setPosition(double position)
   {
      if (position < min)
      {
         this.position = min;
      }
      else if (position > max)
      {
         this.position = max;
      }
      else
      {
         this.position = position;
      }
   }

   /**
    * Return maximum value.
    * 
    * @return  see above.
    */
   public double getVisibleRange()
   {
      return visibleRange;
   }

   /**
    * Set visible range value.
    * 
    * @param   range
    *          New visible range value.
    */
   public void setVisibleRange(double range)
   {
      this.visibleRange = range;
   }

   /**
    * Get scroll bar orientation.
    * 
    * @return  see above.
    */
   public Orientation orientation()
   {
      return orientation;
   }
   
   /**
    * This method is invoked when scroll pane is scrolled.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void scroll(ScrollEvent event)
   {
      ScrollDirection direction = event.direction();
      Point limit = event.limit();

      // update scroll bars
      if (ScrollDirection.left(direction) || ScrollDirection.right(direction))
      {
         if (orientation == Orientation.HORIZONTAL)
         {
            position = Coordinate.toInt(limit.y);
         }
      }
      else if (ScrollDirection.up(direction) || ScrollDirection.down(direction))
      {
         if (orientation == Orientation.VERTICAL)
         {
            position = Coordinate.toInt(limit.y);
         }
      }         
   }
   
   /**
    * Notification on scroll bar button click.
    *  
    * @param   button
    *          Clicked button.
    */
   public void onButtonClick(Position button)
   {    
      ScrollDirection direction = scrollDirection(button);
      double pos = position;
      
      if (direction == ScrollDirection.LEFT || direction == ScrollDirection.UP)
      {
         pos -= step;
      }
      else
      {
         pos += step;         
      }

      notifyScrollListeners(direction, pos);
   }
      
   /**
    * Notification on scroll bar drag
    *  
    * @param   button
    *          Clicked button.
    */
   public void onDrag(Position button, double pos)
   {    
      ScrollDirection direction = scrollDirection(button);
      notifyScrollListeners(direction, pos);
   }
   
   /**
    * Determine scroll direction.
    *  
    * @param   button
    *          Clicked button.
    */
   private ScrollDirection scrollDirection(Position button)
   {    
      if (orientation == Orientation.HORIZONTAL)
      {
         return (button == Position.LEFT) ? ScrollDirection.LEFT : ScrollDirection.RIGHT;
      }
      else
      {
         return (button == Position.LEFT) ? ScrollDirection.UP : ScrollDirection.DOWN;
      }
   }

   /**
    * Attach this widget to the specified container.
    * 
    * @param    parent
    *           The parent container.
    */
   public abstract void attach(AbstractContainer<GuiOutputManager> parent);

   /**
    * Widget drawing method.
    */
   @Override
   public abstract void draw();

   /**
    * Set the step for this scroll-bar.
    * 
    * @param    step
    *           The step.
    */
   public void setStep(double step)
   {
      this.step = step;
   }
   
   /**
    * Get the step of this scroll-bar.
    * 
    * @return   The {@link #step}.
    */
   public double getStep()
   {
      return step;
   }
   
   /**
    * Notify scroll listeners about event.
    * 
    * @param   direction
    *          Scroll direction..
    * @param   pos
    *          New scroll bar position.
    */
   private void notifyScrollListeners(ScrollDirection direction, double pos)
   {
      // notify listeners only if is on range
      if (pos >= min && pos <= max)
      {
         setPosition(pos);
         Point limit = new Point(min, position);
         ScrollEvent event = new ScrollEvent(this, direction, limit);
         
         for (ScrollListener listener : scrollListeners)
         {
            listener.scroll(event);
         }
      }            
   }   
}
