/* 
** Module   : List.java
** Abstract : drop-down list helper implementation for CHUI
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050214   @19831 Creation based on CHARVA JList class.
**                           Necessary modification to get compiled
** 002 EVL 20050523   @21257 Eat the mouse events.
** 003 EVL 20050603   @21405 Changing the inner variable names to sync
**                           with the rest of the project and preparing
**                           to javadoc processing;
** 004 EVL 20050607   @21435 Syntax for comments checking.
** 005 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 006 SIY 20050831   @22507 Refactoring (see @@22394-22409). 
** 007 SIY 20050907   @22641 Fixed use of @inheritDoc tag in javadoc.
** 008 SIY 20050926   @22843 Selected proper attribute container, using 
**                           ColorTable for drawing.
** 009 SIY 20060209   @24457 Cleanups and updates to accommodate changes in
**                           xxxConfig classes, handling of empty list.
** 010 SIY 20060209   @24497 Added getCurrentRow() method. 
** 011 NVS 20060227   @24865 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 012 EVL 20060810   @28543 Adding ability to search the item in combo
**                           box widget as reaction for alphabetical key
**                           pressed when the DropDown is opened.
** 013 SIY 20060824   @28854 Removed getP2JType() method.
** 014 GES 20061107   @30999 Removed dead code and made fixes to key
**                           event processing.
** 015 NVS 20061204   @31555 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 016 SIY 20061212   @31672 Fixed drawing of the drop-down list. Cleanups
**                           and refactoring.
** 017 SIY 20080411   @37938 Fixed drawing of the drop-down list.
** 018 NVS 20080501   @38191 Fixed drawing of the drop-down list one more
**                           time, this time for the empty list. Added
**                           drawCaret() method.
** 019 GES 20090424   @41964 Import change.
** 020 CA  20100507   @44849 Added getBox(), to retrieve the drop-down to
**                           which the list is associated with.
** 021 SIY 20100619          Switching to own implementation of character UI.
** 022 SIY 20100804          Changes in imports, some refactorings.
** 023 SIY 20101122          Switching from KeyEvent to KeyInput.
** 024 SIY 20110112          Changes related to new color handling scheme.
** 024 SIY 20111121          Moved from CHUI and renamed in order to avoid 
**                           conflicts with java.util.List
** 025 SIY 20111202          Refactored and moved from CHUI package.
** 026 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 027 HC  20140911          GUI coordinate support - data types refactoring.
** 028 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 029 EVL 20150517          Adding api to set up the top row index.
*/

package com.goldencode.p2j.ui.client;


import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A component that allows the user to select one or more objects from a
 * list.
 * <p>
 * This class does not provide scrolling capability.  Instances must be
 * inserted into a {@link ScrollPane} to provide scrolling.
 * <p>
 * This class is intended to only be used as the container of the actual
 * "contents" inside the drop-down portion of a {@link ComboBox}. For the
 * general purpose selection list please use the {@link SelectionList} class.
 */
public abstract class ScrollableList<O extends OutputManager<?>>
extends DefaultList<String, O>
{
   /** ComboBox instance which this instance services. */
   protected ComboBox<O> box = null;

   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   public ScrollableList(ComboBox<O> box, ListModel<String> model)
   {
      super(model);
      allowDeselect(false);
      this.box = box;
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public abstract void draw();
   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public abstract void drawCaret();
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   protected abstract void setTopRow(int index);
   
   /**
    * Get item text at specified index.
    * 
    * @param   i
    *          The item index.
    *          
    * @return  The item text.
    */
   protected String getItemText(int i)
   {
      return model().get(i).toString();
   }

   /**
    * Intercept key events and handle some special cases which are not
    * handled by base class.
    * 
    * @param    kvt
    *           Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput kvt)
   {
      int key = kvt.actionCode();
      
      // disable tab and back-tab (we don't give the parent classes a chance
      // to operate on these since they are disabled in Progress)
      if (key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB)
      {
         kvt.consume();
         screen().beep();
         return;
      }

      // END-ERROR action exits the drop down without making any change
      if (key == Keyboard.KA_END_ERROR)
      {
         return;
      }
      
      // Implementation of the item searching by first char. Launches for non
      // functional keys - alphanumerical and for existed combo-box.
      if (kvt.keyCode() < 256 && box != null)
      {
         char keyChar = kvt.keyChar();
         // Only letter or digit is interesting here
         if (keyChar != 0 && Character.isLetterOrDigit(keyChar))
         {
            int newIndx = -1; // We have not yet found new item by default
            ListModel<String> model = model();
            int listSize = model.size();
            // Scan complete list model
            for (int i = 0; i < listSize; i++)
            {
               // Retrieve next item in String representation
               String item = model.get(i).toString();
               // Check equality only for the first char but
               // ignoring the letter case
               if (Character.toUpperCase(item.charAt(0)) ==
                   Character.toUpperCase(keyChar))
               {
                  // Found the candidate, no need to continue search
                  newIndx = i;
                  break;
               }
            }
            // We found the new item to highlight and it is not the same as
            // the current one
            if (newIndx >= 0 && newIndx != currentRow())
            {
               // select it and ensure it will be visible in the list
               ensureRowVisible( newIndx );
               kvt.consume();
               return;
            }
         }
      }
      
      super.processKeyEvent(kvt);
      
      if (key == Keyboard.SE_LEAVE)
      {
         // ignore
      }
      else
      {
         // bypass all further processing
         kvt.consume();
      }
   }

   /**
    * Set the number of columns INSIDE the list.
    */
   @Override
   public void setColumns(double columns)
   {
      // add 2 for the size of the border
      super.setColumns(columns + 2);
   }

   /**
    * Get current cursor row.
    * 
    * @return   The current cursor row.
    */
   public int getCurrentRow()
   {
      return currentRow();
   }
   
   /**
    * Get the combo-box instance to which this drop-down list is attached to.
    * 
    * @return   the {@link ComboBox} instance to which this drop-down 
    *           list is attached to.
    */
   public ComboBox<O> getBox()
   {
      return box;
   }
}
